#!/usr/bin/env python
#-*- coding: utf-8 -*-

"""
    =======
    `Synudic` --- Main module
    =======

    This module contains all classes representing
    a Synudic Sketch, which include functions
    to generate Yices input to solve the corresponding
    SLP synthesis problem.

    This module also contains the main Synudic program.

    Adria Gascon, 2015. 
    Modified by Brent Carmer, 2015.
"""
import os
import sys
import subprocess

import sexpParser
import pyparsing
import itertools
import argparse
import collections
import re
import operator
import copy
import time

def isList(expr):
    return isinstance(expr, pyparsing.ParseResults) or isinstance(expr, list)

OT=False
LINE_TYPE_NAME = 'lineType'
FUNC_TYPE_NAME = 'funcType'
MAX_NUM_BLK_INPUTS = 3

def stripComments(line):
    # ignore comments as everything after ';' in a line
    m = re.search("([^;]*);.*", line)
    if m: line = m.group(1)
    return line

def expr2pstr(e, level = 0):# {{{
    if isinstance(e, str):
        return '  '*level+e
    elif isinstance(e, int):
        return '  '*level+str(e)
    elif isList(e) and all(map(lambda x:isinstance(x, str) or isinstance(x, int), e)):
        if len(e) == 0: return ''
        return '  '*level+'({0})'.format(' '.join(map(str, e)))
    else:
        if len(e) == 0: return ''
        maybe = '  '*level+expr2str(e)
        if len(maybe) <= 120:
            return maybe
        else:
            l = [ expr2pstr(x, level + 1) for x in e[1:] ]
            return '  '*level+'({0}\n{1}\n'.format(expr2str(e[0]),'\n'.join(l))+'  '*level+')'
# }}}
def expr2str(e):# {{{
    if isinstance(e, str):
        return e
    elif isinstance(e, int):
        return str(e)
    else:
        l = [ expr2str(x) for x in e ]
        return '({0})'.format(' '.join(l))
# }}}
def get_block_line_name(block_name, line_num):# {{{
    return '{0}_{1}'.format(block_name, line_num +1)
# }}}
def get_block_from_line_name(line_name):# {{{
    m = re.match('(\S+)_[0-9]+', line_name)
    assert m, line_name
    return m.group(1)
# }}}
def get_block_input_name(block_name, line_num, input_num):# {{{
    return '{0}_arg_{1}'.format(get_block_line_name(block_name, line_num), input_num)
# }}}
def get_block_input_line_name(block_name, line_num, input_num):# {{{
    return '{0}_arg_{1}'.format(get_block_line_name(block_name, line_num), input_num)
# }}}
def get_block_input_val_name(block_name, line_num, input_num):# {{{
    return '{0}_arg_val_{1}'.format(get_block_line_name(block_name, line_num), input_num)
# }}}
def get_block_interp_val_name(block_name, interp_name, line_num): #{{{
    return '{0}_{1}'.format(get_block_line_name(block_name, line_num), interp_name)
# }}}
def get_block_dom_val_name(block_name, line_num):# {{{
    return '{0}_dom'.format(get_block_line_name(block_name, line_num))
# }}}
def get_block_typ_val_name(block_name, line_num):# {{{
    return '{0}_typ'.format(get_block_line_name(block_name, line_num))
# }}}
def get_line_interp_val_name(interp_name, line_name):# {{{
    return '{0}_{1}'.format(line_name, interp_name)
# }}}
def get_line_typ_val_name(line_name):# {{{
    return '{0}_typ'.format(line_name)
# }}}
def get_line_dom_val_name(line_name):# {{{
    return '{0}_dom'.format(line_name)
# }}}
def get_block_function_name(block_name, line_num):# {{{
    return '{0}_func'.format(get_block_line_name(block_name, line_num))
# }}}
def get_exp_dict_from_file(filename, param_dict):# {{{
    try:
        with open(filename, "r") as f:
            string = ""
            for line in f:
                line = stripComments(line)
                # textual replacement of params
                if not re.search("parameters", line):
                    for pname, pval in param_dict.items():
                        pat = "\\b{0}\\b".format(pname)
                        line = re.sub(pat, str(pval), line)
                string += line
            sexpr = sexpParser.sexp.parseString(string, parseAll=True)
        sexpr_list = sexpr.asList()
    except pyparsing.ParseFatalException, pfe:
        print "ParsingError:", pfe.msg
        print pfe.markInputline('^')
    exp_dict = {}
    for i in sexpr_list[0]:
        if isinstance(i,str): continue
        if i[0] in ['primal', 'dual']:
            key = list(itertools.takewhile(lambda x: isinstance(x, str) and x.islower(), i))
            flags = filter(lambda x: isinstance(x, str) and x.isupper(), i)
            val = list(itertools.dropwhile(lambda x: isinstance(x, str), i))
            exp_dict[" ".join(key)] = [i[0], key[1], flags] + val
        else:
            exp_dict[i[0]] = i[1:]
    try:
        assert set(exp_dict['parameters']) == set(param_dict.keys())
    except AssertionError, e:
        print "Error: declared parameters must match those provided on the command line"
        raise e
    return exp_dict
# }}}

def gen_arg_accessor(line_names):# {{{
    s  = "(define arg::(-> lineType int lineType)\n"
    s += "  (lambda (x::lineType i::int)\n"
    for name in line_names:
        s += "    (if (= x {0})\n".format(name)
        for i in range(MAX_NUM_BLK_INPUTS):
            s += "      (if (= i {0}) {1}_arg_{0}\n".format(i, name)
        s += "      " + name
        for i in range(MAX_NUM_BLK_INPUTS):
            s += ")"
        s += "\n"
    s += "    " + name
    for name in line_names:
        s += ")"
    s += "))"
    return s
# }}}

class FunctionConstraint:# {{{
    """
    A Synudic constraint, such as [f [blk3 -] [blk1 blk2 blk3 -]],
    representing all possible expressions f(x, y), where
    x is a variable computed in either blk3 or the current block
    and y is a variable computed in either blk1, blk2, blk3, or the
    current block. We call this a function constraint because
    f is fixed.
    """
    def __init__(self, function, param_constraint):
        self.function = function
        self.param_constraint = param_constraint
        assert len(param_constraint) == function.arity

    def to_yices(self, block, line_number, interp):
        """
        :param block: The block containing the constraint
        :param line_number: The line number containing the constraint
        :interp: the interpretation under consideration
        """
        line_name = block.line_names[line_number]
        line_func_name = block.line_func_name_dict[line_name]
        invalid_lines_in_block = block.line_names[line_number:]
        filtered_param_constraint = map(lambda x:filter(lambda y:not y in invalid_lines_in_block, x), self.param_constraint)
        input_lines_options = itertools.product(*filtered_param_constraint)
        input_lines_formals = block.line_input_line_names_dict[line_name]
        line_interp_name = "{0}_{1}".format(line_name, interp.name)
        input_lines_options_expr = []
        for actuals in input_lines_options:
            assignment = zip(input_lines_formals[:self.function.arity], actuals)
            assignment_expr = ['(= {0} {1})'.format(x,y) for (x,y) in assignment]
            actuals_types = map(lambda x: get_line_interp_val_name(interp.name, x), actuals)
            fname = self.function.get_interp_name(interp.name)
            if interp.primal:
                type_constraint = '(= ({0} {1}) {2})'.format(fname, ' '.join(actuals_types), line_interp_name)
            else:
                type_constraint = '({0} {1} {2})'.format(fname, ' '.join(actuals_types), line_interp_name)
            input_lines_options_expr.append(['and'] + assignment_expr + [type_constraint])
        if not input_lines_options_expr: return []
        input_lines_options_expr = ['or'] + input_lines_options_expr
        line_func_constraint = ('(= {0} {1})'.format(line_func_name, self.function.name))
        block_yices_expr = ['and'] + [line_func_constraint] + [input_lines_options_expr]

        return block_yices_expr
# }}}
class BlockDecl:# {{{
    """
    A block in a synudic sketch
    """

    def __init__(self, name, length, constraints, previous_blocks, signature):
        self.name = name
        self.length = int(length)
        self.previous_blocks_dict = { b.name : b for b in previous_blocks }
        self.line_names = [get_block_line_name(self.name, i) for i in range(self.length)]
        self.signature = signature
        if constraints[0][0] == 'input':
            self.isinput = True
            assert len(self.line_names) == 1
            length = len(constraints[0])
            if length == 2:
                if constraints[0][1].find('::') != -1:
                    init_val = constraints[0][1][constraints[0][1].find('::')+2:]
                    self.constraints = ['(= {0} {1})'.format(get_line_typ_val_name(self.line_names[0]), init_val)]
                else:
                    self.constraints = []
            elif length == 3:
                init_val = expr2pstr(constraints[0][2])
                self.constraints = ['(= {0} {1})'.format(get_line_typ_val_name(self.line_names[0]), init_val)]
            else:
                raise Exception("unknown length constraints")
        else:
            self.isinput = False
            self.constraints = []
            for c in constraints:
                c_function_name = c[0]
                c_function = self.signature.get_function_symbol(c_function_name)
                params_constraint = []
                for p in c[1:]:
                    param_constraint = []
                    for block_name in p:
                        if block_name != '-':
                            param_constraint += self.previous_blocks_dict[block_name].line_names
                        else:
                            param_constraint += self.line_names
                    params_constraint.append(param_constraint)

                constraint = FunctionConstraint(c_function, params_constraint)
                self.constraints.append(constraint)
                assert len(params_constraint) == c_function.arity, \
                '{0} != {1}'.format(len(params_constraint), c_function.arity)

        self.line_input_line_names_dict = dict(
            [(self.line_names[i], tuple([get_block_input_line_name(self.name, i, j)
                        for j in range(MAX_NUM_BLK_INPUTS)]))
             for i in range(self.length)]
            )

        self.line_func_name_dict = dict(
            [(self.line_names[i], get_block_function_name(self.name, i))
             for i in range(self.length)])

    def get_block_vars(self):
        variables = [] # [(x, type)]
        if not self.isinput:
            for l in self.line_names:
                for i in self.line_input_line_names_dict[l]:
                    variables.append((i, LINE_TYPE_NAME))
            for n,f in self.line_func_name_dict.iteritems():
                variables.append((f, FUNC_TYPE_NAME))
        return variables

    def get_constraint(self, interp):
        """
        :return: An expression containing the Yices encoding of the block
        """
        constraint = []
        variables = []
        for i in range(self.length):
            val = get_block_interp_val_name(self.name, interp.name, i) 
            variables.append((val, interp.name))

        for i, l in enumerate(self.line_names):
            expr = []
            for c in self.constraints:
                if isinstance(c, str):
                    expr.append(c)
                else:
                    s = c.to_yices(self, i, interp)
                    if s: expr.append(s)
            if expr:
                expr.insert(0, 'or')
                constraint.append(expr)

        if constraint: constraint.insert(0, 'and')
        return [variables, constraint]
# }}}
class Signature:# {{{
    """
    The Signature in a synudic sketch

    * Attributes:
        * function_symbol_list
        * function_name2function_symbol
    """
    def __init__(self, function_symbol_list):
        self.function_symbol_list = function_symbol_list
        self.function_name2function_symbol = dict([(x.name, x) for x in self.function_symbol_list])

    def add_symbol(self, s):
        self.function_symbol_list.append(s)
        self.function_name2function_symbol[s.name] = s

    def get_constants(self):
        return [f for f in self.function_symbol_list if f.arity == 0]

    def get_function_names(self):
        return [x.name for x in self.function_symbol_list]

    def get_functions(self):
        return self.function_name2function_symbol.values()

    def get_function_symbol(self, name):
        return self.function_name2function_symbol[name]
# }}}
class FunctionSymbol:# {{{
    """
    The Signature in a synudic sketch

    * Attributes:
        * name
        * arity
        * is_ac (associative, commutative)
        * interp_decl
        * interp_names
    """
    def __init__(self, name, arity, is_ac = False):
        self.name = name
        self.arity = arity
        self.is_ac = is_ac
        self.interp_decl = {}
        self.interp_names = {}

    def get_input_filter(self):
        """
        :returns:
        A function that takes a set of input arguments and returns a
        filtered list with equivalent inputs removed.
        This function is used to avoid
        considering equivalent possibilities
        for commutative operations.
        """
        def id(l):
            return l
        def filter_commutative(l):

            if len(l) == 0: return l
            if len(l[0]) == 2:
                redundant = filter(lambda (x,y): (y,x) in l and y < x, l)
                return filter(lambda x:not x in redundant, l)
            return l
        if self.is_ac:
            return filter_commutative
        else:
            return id
    
    def set_interp_decl(self, interp_name, func_interp_name):
        self.interp_names[interp_name] = func_interp_name

    def get_interp_name(self, interp_name):
        return self.interp_names[interp_name]
# }}}
class Interp:# {{{
    """
    A primal or dual interpretation

    * Attributes:
        * type_decls
        * self
        * primal
        * params
        * signature # global signature
        * type_decls
        * aux_fun_decls
        * ensure_expr
        * block_constraints
    """
    def __init__(self, name, sexp, signature):
        self.primal = sexp[0] == 'primal'
        self.name   = sexp[1]
        self.params = sexp[2]
        self.signature = signature
        self._read_sexp(sexp[3:])
        self.block_constraints = []
        self.interp_decls = []
        self.var_decls = {}
        self.line_names = []
        self.input_line_names = []
        self.block_names = {}
        return None

    def _read_sexp(self, sexp):
        assert len(sexp) <= 2
        decls = []
        ensure = []
        for expr in sexp:
            if expr[0] == 'decls':
                decls = expr[1:]
            elif expr[0] == 'ensure':
                ensure = expr[1:]
                if len(ensure) == 1:
                    ensure = ensure[0]

        assert not self.primal or decls

        type_decls = []
        fun_decls = []
        fun_decls_names = []
        aux_fun_decls = []
        for d in decls:
            if d[0] == 'define-type':
                type_decls.append(expr2str(d))
            elif d[0] == 'define':
                fname = re.match('[^:\s]*', d[1]).group()
                if isinstance(d[2], str) and re.match('^[\s:]+$', d[2]): # parser separates colons from words
                    d.pop(2)                                             # adjust for that
                if fname in self.signature.get_function_names():
                    # name mangle this function
                    fname_mod = "{0}-{1}".format(self.name, fname)
                    d[1] = fname_mod + " ::"
                    # then we'll have to customize internal function names for this interpretation
                    def rec(expr):
                        if isinstance(expr, str):
                            if expr in self.signature.get_function_names():
                                return "{0}-{1}".format(self.name, expr)
                            else:
                                return expr
                        elif isinstance(expr, list):
                            return [ rec(e) for e in expr ]
                        else:
                            return expr
                    d = rec(d)
                    sym = self.signature.get_function_symbol(fname)
                    sym.set_interp_decl(self.name, fname_mod)
                    fun_decls.append(expr2str(d))
                    fun_decls_names.append(fname)
                else:
                    aux_fun_decls.append(expr2str(d))
            elif d[0] == 'constraint':
                self.parametric_constraint = d
            else:
                raise ValueError("Not supported: {0}".format(d[0]))

        if not 'GLOBAL' in self.params:
            ns = set(fun_decls_names) 
            ss = set(self.signature.get_function_names())
            assert ns == ss,\
                    "missing function definitions in interp {0}: {1}".format(self.name, " ".join(ss.difference(ns)))

        self.type_decls = type_decls
        self.fun_decls = fun_decls
        self.aux_fun_decls = aux_fun_decls
        self.ensure_expr = ensure

    def get_ensure_expr(self):
        if self.ensure_expr:
            res = self._expand_loops(self.ensure_expr)
            return self._replace_macros(res)
        else:
            return ""

    def add_blocks(self, blocks):
        # Block constraints
        for b in blocks:
            self.block_names[b.name] = []
            if b.isinput:
                self.input_line_names += b.line_names
            else:
                self.line_names += b.line_names # need to know line_names for generating parametric constraints
                self.block_names[b.name].extend(b.line_names)
            line_vars = b.get_block_vars()
            if line_vars: 
                self.var_decls.update(dict(line_vars))
            if not "GLOBAL" in self.params:
                [block_interp_decls, block_constraint] = b.get_constraint(self)
                self.interp_decls += block_interp_decls
                if block_constraint: 
                    self.block_constraints.append(block_constraint)

    def gen_parametric_constraint(self):
        c = self.parametric_constraint
        # definine the number of levels
        assert len(c[1]) == 1, "unknown syntax: {0}".format(expr2str(c))
        [nlevels] = c[1]
        # need to replace this-level and next-level
        def write_levels(expr, level):
            if isinstance(expr, list):
                if expr[0] == "this-level":
                    assert len(expr) == 2, "syntax error: {0}".format(expr)
                    expr[0] = self.parametric_level_accessor
                    expr.append(level)
                    return expr
                elif expr[0] == "next-level":
                    assert len(expr) == 2, "syntax error: {0}".format(expr)
                    expr[0] = self.parametric_level_accessor
                    expr.append(level+1)
                    return expr
                else:
                    return [write_levels(x, level) for x in expr]
            else:
                return expr
        cs = []
        for lvl in range(nlevels-1):
            cs.append(";; level {0}".format(lvl+1))
            expr = write_levels(copy.deepcopy(c[2]), lvl)
            expr = self._expand_loops(expr)
            expr = self._replace_macros(expr)
            cs.append(expr)
        cs.insert(0, "and ;; expansion of of {0} constraint".format(self.name))
        return cs
        
    # unrolls all foreach's in expr
    # returns a bool indicating that the result must be spliced into its parent
    def _expand_loops(self, expr):
        def recurse(expr):
            if isinstance(expr, list):
                if expr[0] == "foreach":
                    res = self._unroll_foreach(expr)
                    res.insert(0, 'and')
                    return (res, False)
                elif expr[0] == "forsome":
                    res = self._unroll_foreach(expr)
                    res.insert(0, 'or')
                    return (res, False)
                elif expr[0] == "forsplice":
                    res = self._unroll_foreach(expr)
                    return (res, True)
                else:
                    res = []
                    for e in expr:
                        (expr, splice) = recurse(e)
                        if splice:
                            res += expr
                        else:
                            res.append(expr)
                    return (res, False)
            else:
                return (expr, False)
        (res, splice) = recurse(expr)
        return res

    def _unroll_foreach(self, f):
        assert f[0] in ["foreach", "forsplice", "forsome"]
        assert len(f) == 3, "foreach expr must have 3 terms"
        assert self.line_names
        foreach_vars = []
        choices = []
        for v in f[1]:
            assert len(v) > 1, v
            foreach_vars.append(v[0])
            c = []
            for e in v[1:]:
                if e == "lines":
                    c.extend(self.line_names)
                elif e == "inputs":
                    c.extend(self.input_line_names)
                elif e in self.block_names:
                    c.extend(self.block_names[e])
                else:
                    raise AssertionError, "unknown keyword: " + e
            choices.append(c)
        def recurse(expr, assn):
            if isinstance(expr, str) or isinstance(expr, int):
                if expr in foreach_vars: 
                    i = foreach_vars.index(expr)
                    return assn[i]
                else:
                    return expr
            elif isinstance(expr, list):
                expr = self._expand_loops(expr)
                res = [ recurse(x, assn) for x in expr ]
                return res
        # generate possible assignments to each line var
        # finding the cartesian product is sketchy in python
        assignments = itertools.product(*choices) 
        res = [ recurse(f[2], assn) for assn in assignments ]
        return res

    # replace (func LINE), (level STUFF n), (line NAME NUM) with concrete vars
    def _replace_macros(self, expr):
        assert self.line_names, "Interp {0} must known line names!".format(self.name)
        def recurse(expr):
            if isinstance(expr, list):
                if expr[0] == "func":
                    assert len(expr) == 2, expr
                    s = recurse(expr[1])
                    # assert s in self.line_names, s
                    return s + "_func"
                elif expr[0] == self.name:
                    if len(expr) == 2: # occurs in foreach loops: eg (foreach ((x lines)) (typ x)) <- x is a lineType
                        assert expr[1] in self.line_names+ self.input_line_names
                        return "{0}_{1}".format(expr[1], self.name)
                    if len(expr) == 3:
                        s = recurse(expr[1])
                        assert isinstance(expr[2], int), expr
                        line_name = s + "_" + str(expr[2])
                        assert line_name in self.line_names + self.input_line_names, line_name + " unknown"
                        return "{0}_{1}".format(line_name, self.name)
                elif expr[0] == "level" and len(expr) == 3:
                    assert self.parametric_level_accessor, "no level accessor defined"
                    assert isinstance(expr[2], int), expr
                    s = recurse(expr[1])
                    assert s in self.line_names + self.input_line_names, s
                    return [ self.parametric_level_accessor, s, expr[2] - 1 ]
                elif expr[0] == "line" and len(expr) == 3:
                    assert isinstance(expr[2], int), expr
                    s = expr[1] + "_" + str(expr[2])
                    assert s in self.line_names + self.input_line_names, s
                    return s 
                else:
                    return [ recurse(x) for x in expr ]
            else:
                return expr
        return recurse(expr)

    def gen_parametric_level_accessor(self):
        assert "PARAMETRIC" in self.params
        assert self.line_names and self.parametric_variables and self.parametric_constraint
        lines = self.line_names + self.input_line_names
        nlevels = self.parametric_constraint[1][0]
        func_name = "{0}-get-level".format(self.name)
        s = "(define {0}::(-> lineType int {1})\n".format(func_name, self.name)
        s += "  (lambda (x::lineType lvl::int)\n"
        for line in lines:
            s += "    (if (= x {0})".format(line)
            for l in range(nlevels):
                s += "\n      (if (= lvl {0}) {1} ".format(l, self.parametric_variables[line][l])
            s += "\n      false" # final clause in levels switch
            for l in range(nlevels-1):
                s += ")"
            s += ")\n"
        s += "     false" # final clause in line switch
        for line in lines:
            s += ")"
        s += "\n  )\n" # lambda
        s += ")"   # define
        self.parametric_level_accessor = func_name
        return [s]

    def get_parametric_variable_name(self, level, line_name):
        return "{0}_{1}_lvl_{2}".format(self.name, line_name, level)

    def gen_parametric_variables(self):
        assert "PARAMETRIC" in self.params
        assert self.line_names, self.parametric_constraint
        self.parametric_variables = {} 
        nlevels = self.parametric_constraint[1][0]
        for line in self.line_names + self.input_line_names:
            self.parametric_variables[line] = []
            for lvl in range(nlevels):
                self.parametric_variables[line].append(self.get_parametric_variable_name(lvl, line))
        new_vars = sum(self.parametric_variables.values(), []) # concatenate
        decls = [ "(define {0} :: {1})".format(x, self.name) for x in new_vars ]
        return decls

    def gen_aux_defns(self):
        content = []
        content += self.aux_fun_decls
        if "PARAMETRIC" in self.params:
            content += ["\n;; Parametric declarations for {0} interpretation".format(self.name)]
            content += self.gen_parametric_variables()
            content += ["\n;; Parametric level accessor for {0} interpretation".format(self.name)]
            content += self.gen_parametric_level_accessor()
        return "\n".join(content)

    def gen_type_decls(self):
        return "\n".join(self.type_decls)

    def get_var_decls(self):
        return self.var_decls.copy()

    def gen_constraints(self):
        content = []
        content.append("\n;; {0} constraints".format(self.name))
        if self.primal:
            content.append("(assert\n  (forall (")
            for (d,t) in self.interp_decls:
                content.append("    {0} :: {1}".format(d,t))
            content.append("  )\n  (=>\n    (and")
            for c in self.block_constraints:
                content.append(expr2pstr(c, 3))
            content.append("    )") # and
            content.append(expr2pstr(self.get_ensure_expr(), 2))
            content.append("  )\n  )\n)") # =>, forall, assert

        elif not self.primal and not "GLOBAL" in self.params: # dual non-parametric interp
            for (d,t) in self.interp_decls:
                content.append("(define {0} :: {1})".format(d,t))
            content.append("(assert\n  (and")
            for c in self.block_constraints:
                content.append(expr2pstr(c, 2))
            content.append("    ;; {0} ensure expression".format(self.name))
            content.append(expr2pstr(self.get_ensure_expr(), 2))
            content.append("  )\n)") # assert

        elif not self.primal and "GLOBAL" in self.params and "PARAMETRIC" in self.params:
            content.append("\n;; {0} global parametric constraints {1}".format(self.name, "")) #'{'*3))
            content.append("(assert")
            content.append("  (and")
            content.append("    ;; parametric constraint")
            content.append(expr2pstr(self.gen_parametric_constraint(), 2))
            content.append("    ;; ensure expr")
            content.append(expr2pstr(self.get_ensure_expr(), 2))
            content.append("  )")
            content.append(") ;; {0}".format("")) #'}'*3))

        return "\n".join(content)

# }}}
class Sketch:# {{{
    """
    A synudic sketch

    * Attributes:
        * blocks
        * signature
        * interps
        * existsforall

    """
    def __init__(self,
                 blocks,
                 signature,
                 interps,
                 existsforall):

        self.blocks = blocks
        self.signature = signature
        self.interps = interps
        self.existsforall = existsforall

    @classmethod
    def from_file(cls, filename, param_dict):
        """
        """
        exp_dict = get_exp_dict_from_file(filename, param_dict)
        try:
            signature = Signature([FunctionSymbol(x[0], x[1], x[2] == 'ac') for x in exp_dict['library']])
        except IndexError as e:
            signature = Signature([FunctionSymbol(x[0], x[1]) for x in exp_dict['library']])

        interps = []
        existsforall = False
        for k, v in exp_dict.iteritems():
            if re.match('^primal|dual', k):
                i = Interp(k, v, signature)
                interps.append(i)
                if i.primal:
                    existsforall = True

        blocks = []
        for (block_name, block_length, block_desc) in  exp_dict['blocks']:
            b = BlockDecl(block_name, block_length, block_desc, blocks, signature)
            blocks.append(b)

        for interp in interps:
            interp.add_blocks(blocks)

        return cls(blocks, signature, interps, existsforall)

    def to_yices(self):
        content = [';; Line type declarations']
        all_line_names = [' '.join(b.line_names) for b in self.blocks]
        line = '(define-type {0} (scalar {1}))'.format(LINE_TYPE_NAME, ' '.join(all_line_names))
        content.append(line)

        content.append('\n;; Function type declarations')
        all_func_names = [f.name for f in self.signature.function_symbol_list] + ["input"]
        line = '(define-type {0} (scalar {1}))'.format(FUNC_TYPE_NAME, ' '.join(all_func_names))
        content.append(line)

        content.append('\n;; Input declarations')
        all_inputs = set()
        for i in self.interps:
            for l in i.input_line_names:
                all_inputs.add(l)
        for l in all_inputs:
            line = '(define {0}_func::funcType input)'.format(l)
            content.append(line)

        content.append("\n;; Interpretation type declarations")
        for interp in self.interps:
            content.append(interp.gen_type_decls())

        content.append("\n;; Function definitions")
        for interp in self.interps:
            content.append(interp.gen_aux_defns())
            if not "GLOBAL" in interp.params:
                for f in interp.fun_decls:
                    content.append(f)

        # there will be overlap in var decls, since we initialize multiple
        # interpretations with the same blocks. So we have to keep track of them
        # and make sure not to redefine existing variables. Hence the dictionary.
        content.append("\n;; Line, func, and arg definitions")
        var_decls = {}
        for interp in self.interps:
            var_decls.update(interp.get_var_decls())
        for name, ty in var_decls.iteritems():
            s = "(define {0} :: {1})".format(name, ty)
            content.append(s)

        content.append("\n;; Argument accessor")
        content.append(gen_arg_accessor(self.interps[0].line_names))

        # generate assertions
        for interp in self.interps:
            content.append(interp.gen_constraints())

        if self.existsforall:
            content.append('(ef-solve)')
        else:
            content.append('(check)')
            content.append('(show-model)')

        return '\n'.join(content)

    def show_result(self, model, block, types=None):
        program_lines = {}
        processed_blocks = []
        for i, l in list(enumerate(block.line_names)):
            try:
                line_func_name = model[get_block_function_name(block.name, i)]
            except KeyError as e:
                # b is an input block
                assert len(b.line_names) == 1
                #print '{0}'.format(l)
                continue
            f = self.signature.get_function_symbol(line_func_name)
            line_args = [model[arg] for arg in [get_block_input_line_name(b.name, i, j) for j in range(f.arity)]]
            program_lines[l] = (line_func_name, tuple(line_args))
            print '{0} = {1}({2})'.format(l, line_func_name, ' '.join(line_args))

            def printOT(t):
                for m in ['m1', 'm2']:
                    print '  ' + m + ':' , 
                    if m in t:
                        for a in ['a1', 'a2', 'a3']:
                            if a in t[m]:
                                if a == 'a2': print '+',
                                elif a == 'a3': print '-',
                                print "".join(t[m][a]),
                print

            if types and l in types:
                for i in types[l].keys():
                    if OT:
                        printOT(types[l][i])
                    else:
                        types[l][i].sort()
                        print '  {0} :: {1}'.format(i, " ".join(types[l][i]))
            for i in self.interps:
                n = l + "_" + i.name
                if n in model:
                    m = re.match("(.*)_" + i.name, model[n])
                    if m:
                        s = m.group(1)
                        if s != l and s in types and i.name in types[s]:
                            if OT:
                                printOT(types[s][i.name])
                            else:
                                types[s][i.name].sort()
                                print '  {0} :: {1}'.format(i.name, " ".join(types[s][i.name]))
                    else:
                        print '  {0} :: {1}'.format(i.name, model[n])

            print ""


# }}}

def print_usage():# {{{
    print 'The Synudic SLP synthesizer'
    print 'Usage: python synudic.py in_file.sketch p1=v1 p2=v2 p3=v3 ...'
# }}}
def parse_yices_output(filename, existsforall):# {{{
    parsing_model = False
    model = {}
    types = []
    with open(filename, 'r') as f_desc:
        for line in f_desc.readlines():
            if not line.strip(): continue
            if line.strip() == 'unsat':
                return None
            elif line.strip() == 'sat':
                parsing_model = True
                continue
            if parsing_model:
                if existsforall: # the output format is different for the ef solver
                    m = re.match('\s*(\S+)\s*:=\s*(\S+)', line)
                else:
                    m = re.match ('\(= (\S+) (\S+)\)', line)
                if m and line:
                    var, val = m.group(1), m.group(2)
                    model[var] = val
                else:
                    if re.match('.*function.*', line):
                        types.append(line.rstrip())
                    else:
                        types[-1] += line.rstrip()
    if not existsforall:
        model["_types_"] = parse_yices_types(types)
    return model
# }}}

def convert_z3_output(filename, sketch):# {{{
    with open(filename, 'r') as f:
        sat = f.readline().strip()
        if sat != 'sat': return
        s = ""
        for line in f: 
            s += stripComments(line.strip())

    out = ""
    sexp = sexpParser.sexp.parseString(s, parseAll=True)
    mappings = {}
    lineNames = sketch.interps[0].line_names + sketch.interps[0].input_line_names
    for e in sexp[0]:
        if isList(e):
            if len(e) == 4 and e[0] == 'define-fun' and e[1] in sketch.signature.get_function_names() and (not isList(e[3])):
                mappings[e[3]] = e[1]
            if len(e) == 4 and e[0] == 'define-fun' and e[1] in lineNames and (not isList(e[3])):
                mappings[e[3]] = e[1]        

    with open(filename, 'w') as f:
        f.write("sat\n")
        for e in sexp[0]:
            if len(e) == 4 and e[0] == 'define-fun' and e[3] in mappings:
                s = "(= {0} {1})\n".format(e[1], mappings[e[3]])
                f.write(s)
# }}}
def parse_yices_types(typs): #{{{
    model = {}
    for t in typs:
        sexp = sexpParser.sexp.parseString(t, parseAll=False)
        if sexp[0][0] == 'function':
            typeName = sexp[0][2][1]
            m = re.match("(.*)_" + typeName, sexp[0][1])
            lineName = m.group(1)
            if not lineName in model: model[lineName] = {}
            if OT:
                if not typeName in model[lineName]: model[lineName][typeName] = {}
                m = model[lineName][typeName]
                for s in sexp[0][3:]:
                    if s[0] == '=':
                        branch = s[1][1]
                        addend = s[1][2]
                        if not branch in m: m[branch] = {}
                        if not addend in m[branch]: m[branch][addend] = []
                        if s[-1] != 'bot':
                            m[branch][addend].append(s[-1])
                    elif s[0] == 'default' and s[1] != 'bot':
                        for b in ['m1','m2']:
                            for a in ['a1', 'a2', 'a3']:
                                if not b in m: m[branch] = {}
                                if not a in m[b]: m[b][a] = []
                                m[b][a].append(s[1])

            else:
                if not typeName in model[lineName]: model[lineName][typeName] = []
                for s in sexp[0][3:]:
                    if s[0] == "=":
                        x = "[" + " ".join(s[1][1:] + [s[2]]) + "]"
                        model[lineName][typeName].append(x)
                    else:
                        x = "[~" + " ".join(s) + "]"
                        model[lineName][typeName].append(x)
    return model
# }}}
def expr_to_z3(exprs): # {{{
    ret = []
    arrays = []
    for expr in exprs:
        if expr[0] == "define-type":
            ret += z3_define_type(expr, arrays) # may return multiple elements
        elif expr[0] == "define":
            ret.append(z3_define_function(expr, arrays))
        elif expr[0] == 'check':
            ret.append(['check-sat'])
        elif expr[0] == 'show-model':
            ret.append(['get-model'])
        else:
            ret.append(expr)
    return ret

def z3_define_function(expr, arrays):
    fexpr = None
    # different cases occur because "stuff::type" parsing is inconsistent
    m = re.match('^([^:]*)::([^:]*)$', expr[1])
    if m:
        fname = m.group(1)
        if m.group(2):
            ftype = m.group(2)
            if len(expr) == 3:
                fexpr = expr[2]
        else:
            ftype = expr[2]
            if len(expr) == 4:
                fexpr = expr[3]
    elif expr[2] == '::':
        fname = expr[1]
        ftype = expr[3]
        if len(expr) == 5:
            fexpr = expr[4]

    if not fexpr and isinstance(ftype, str):
        ret = ["declare-fun", fname, [], ftype]

    elif fexpr and isinstance(ftype, str) and (isinstance(fexpr, str) or isinstance(fexpr, int)):
        ret = ["define-fun", fname, [], ftype, fexpr]

    elif fexpr and isList(ftype) and ftype[0] == '->' and fexpr[0] == 'lambda':
        args = []
        arrayArgs = []
        for arg in fexpr[1]:
            m = re.match('^([^:]*)::([^:]*)$', arg)
            if m.group(2) in arrays: arrayArgs.append(m.group(1))
            args.append([m.group(1), m.group(2)])
        resultType = ftype[-1]
        body = fexpr[2]
        def insertSelect(expr):
            if isList(expr) and expr[0] in arrayArgs:
                expr.insert(0, 'select')
            if isList(expr):
                for e in expr: insertSelect(e)
        def replaceUpdate(expr):
            if isList(expr) and expr[0] == 'update' and isList(expr[2]):
                rec1 = replaceUpdate(expr[1])
                rec3 = replaceUpdate(expr[3])
                ret = ['store', rec1]
                ret.extend(expr[2])
                ret.append(rec3)
                return ret
            elif isList(expr):
                return [ replaceUpdate(e) for e in expr ]
            else:
                return expr
        insertSelect(body)
        body = replaceUpdate(body)
        ret = ["define-fun", fname, args, resultType, body]
    else:
        ret = []
    return ret

def z3_define_type(expr, arrays):
    ret = []
    if len(expr) == 3 and isinstance(expr[2], pyparsing.ParseResults) and expr[2][0] == 'scalar':
        typeName = expr[1]
        ret.append(["declare-sort", typeName, 0])
        for constName in expr[2][1:]:
            ret.append(["declare-fun", constName, [], typeName])
    if len(expr) == 3 and isinstance(expr[2], pyparsing.ParseResults) and expr[2][0] == '->':
        typeName = expr[1]
        ret.append(["define-sort", typeName, [], ["Array"] + expr[2][1:]])
        arrays.append(typeName)
    elif len(expr[2]) == 3 and isinstance(expr[2], str):
        typeName = expr[1]
        target = expr[2]
        ret.append(["define-sort", typeName, [], target])
    return ret

def z3_capitalize(expr):
    if expr in ['bool', 'int', 'real']:
        return expr.capitalize()
    elif isList(expr):
        return [ z3_capitalize(e) for e in expr ]
    else:
        return expr
# }}}

if __name__ == "__main__":# {{{
    args = sys.argv[1:]
    if len(args) == 0:
        print_usage()
        sys.exit(1)
    sketch_filename = args[0]
    assert os.path.isfile(sketch_filename), 'Error: File {0} does not exist'.format(filename)
    params, param_dict = args[1:], {}
    z3 = False
    for i in params:
        if i == '-ot': 
            OT = True
        elif i == '-z3': 
            z3 = True
        else:
            varval = i.split('=')
            assert len(varval) == 2, 'Err: var=val expected, found {0}'.format(i)
            param_dict[varval[0]] = varval[1]

    sketch = Sketch.from_file(sketch_filename, param_dict)
    yices_file_content = sketch.to_yices()

    yices_filename = 'synth.ys'
    smt_filename = 'synth.z3'
    out_filename = os.path.split(sketch_filename)[0] + '.yout'
    tmp_filename = 'tmp.txt'
    out_file_desc = open(out_filename, 'w')

    assert not (z3 and sketch.existsforall), "z3 converter does not currently support EF"

    with open(yices_filename, "w") as fp:
        if not z3:
            print >> fp, '(set-param ef-max-iters 20000)'
        #print >> fp, str(yices_file_content)
        print >> fp, yices_file_content
        print "Created file {0}".format(yices_filename)

    if z3:
        print "writing z3 file"
        with open(yices_filename, "r") as fp:
            s = ""
            for line in fp:
                if re.search("^\s*$", line): continue
                m = re.search("^([^;]*);.*$", line)
                if m: line = m.group(1)
                s += line
            s = "(" + s + ")"
            sexpr = sexpParser.sexp.parseString(s, parseAll=True)
        with open(smt_filename, "w") as f:
            for eprime in expr_to_z3(sexpr[0]):
                s = expr2pstr(z3_capitalize(eprime),0)
                f.write(s + "\n")

    with open(tmp_filename, 'w' ) as tmp_file_desc:
        if sketch.existsforall:
            yices_command_list = ['yices', '--mode=ef', '--verbosity=10', yices_filename]
        elif z3:
            # yices_command_list = ['z3', '-v:100', smt_filename]
            yices_command_list = ['z3', smt_filename]
        else:
            yices_command_list = ['yices', '--mode=one-shot', '--verbosity=10', yices_filename]
        print 'Running "{0}"'.format(' '.join(yices_command_list))
        start = time.time()
        subprocess.call(yices_command_list, stdout=tmp_file_desc)
        end = time.time()
        print "smt took {0} seconds".format(end - start)

    if z3:
        convert_z3_output(tmp_filename, sketch)
    model = parse_yices_output(tmp_filename, sketch.existsforall)
    if model:
        for b in sketch.blocks:
            print "-- block \"{0}\" --".format(b.name)
            if "_types_" in model:
                sketch.show_result(model, b, model["_types_"])
            else:
                sketch.show_result(model, b)
# }}}
